/*

This is an internal preprocessor for ScriptBasic. This program has to be compiled
to a dll or so and is loaded into the basic interpreter without starting a new
process.

Note however that the implementation of internal preprocessors is experimental
and will radically change in the future in a non-compatible manner.

*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "../conftree.h"
#include "../report.h"
#include "../reader.h"
#include "../basext.h"
#include "../prepext.h"

void CBF_ListLocalVars(char *pszName,
                       void *pSymbol,
                       void *__DUMMY__){
  pSymbolVAR pVAR = pSymbol;

  printf("%s %d\n",pszName,pVAR->Serial);
  }



int MyExecBefore(pExecuteObject pEo){
  printf("PC=%d\n",pEo->ProgramCounter);
  printf("Press enter to continue...");
  getchar();
  return 0;
  }
int MyExecAfter(pExecuteObject pEo){
  return 0;
  }
int MyExecCall(pExecuteObject pEo){
  return 0;
  }
int MyExecReturn(pExecuteObject pEo){
  return 0;
  }


int DLL_EXPORT preproc(pPrepext pEXT,
                       long *pCmd,
                       void *p
  ){

  switch( *pCmd ){
    case PreprocessorLoad:
      *pCmd = PreprocessorContinue;
      printf("Preprocessor loaded\n");
      return 0;

    case PreprocessorReadStart:
      printf("Read starts\n");
      return 0;
    case PreprocessorReadDone0:
    case PreprocessorReadDone1:
    case PreprocessorReadDone2:
    case PreprocessorReadDone3:
      printf("ReadDone%d\n",*pCmd-PreprocessorReadDone0);
      return 0;
#define X(x) case x: printf("%s\n",#x); return 0;
  X(PreprocessorLexInit)
  X(PreprocessorLexDone)
  case PreprocessorLexNASymbol:
    { pLexeme **pLx = p;
      printf("This is NASymbol: %d\n",(**pLx)->value.lValue);
    }
    return 0;
  case PreprocessorLexASymbol:
    { pLexeme **pLx = p;
      printf("This is ASymbol: %d\n",(**pLx)->value.lValue);
    }
    return 0;
  case PreprocessorLexSymbol:
    { pLexeme **pLx = p;
      printf("This is Symbol: %s\n",(**pLx)->value.sValue);
      if( ! strcmp((**pLx)->value.sValue,"skip_this_symbol") ){
        (**pLx)->type = LEX_T_SKIP_SYMBOL;
        }
    }
    return 0;
  case PreprocessorLexMString :
  case PreprocessorLexString :
    { pLexeme **pLx = p;
      printf("This is a string: %s\n",(**pLx)->value.sValue);
    }
    return 0;
  case PreprocessorLexInteger:
    { pLexeme **pLx = p;
      printf("This is an integer number: %d\n",(**pLx)->value.lValue);
    }
    return 0;
  case PreprocessorLexReal:
    { pLexeme **pLx = p;
      printf("This is a real number: %f\n",(**pLx)->value.dValue);
    }
    return 0;
  case PreprocessorLexCharacter:
    { pLexeme **pLx = p;
      printf("This is a character, code: %d\n",(**pLx)->value.lValue);
    }
    return 0;

  case PreprocessorExStart:
    {
      printf("Ex started.\n");
    }
    return 0;
  case PreprocessorExStartLine:
    {
      printf("Ex line started\n");
    }
    return 0;
  case PreprocessorExEnd:
    {
      printf("Ex End\n");
    }
    return 0;
  case PreprocessorExFinish:
    {
      peXobject pEx = p;
      peNODE_l Result = pEx->pCommandList;
      printf("The start node is %d\n",Result->NodeId);
      printf("The number of nodes is %d\n",pEx->NodeCounter);
      pEXT->pPointer-> pEXT->pST->Alloc(sizeof()*pEx->NodeCounter,pEXT->pMemorySegment);
      while( Result ){
        printf("Node %d is %s(%d)\n",Result->NodeId,Result->szFileName,Result->lLineNumber);
        Result = Result->rest;
        }
      printf("Ex finish\n");
    }
    return 0;
  case PreprocessorExStartLocal:
    {
      peXobject pEx = p;
      printf("Preprocessor Start Local: %s\n",pEx->ThisFunction->FunctionName);
    }
    return 0;
  case PreprocessorExEndLocal:
    {
      peXobject pEx = p;
      printf("Preprocessor End Local: %s\n There are %d local variables.\n",
             pEx->ThisFunction->FunctionName,
             pEx->cLocalVariables);
      printf("Listing local variables:\n");
      pEXT->pST->TraverseSymbolTable(pEx->LocalVariables,CBF_ListLocalVars,NULL);
    }
    return 0;

  case PreprocessorExeStart:
    { pExecuteObject pEo = p;
      pEo->pHookers->hook_pointer = pEXT;
      pEo->pHookers->HOOK_ExecBefore = MyExecBefore;
      pEo->pHookers->HOOK_ExecAfter = MyExecAfter;
      pEo->pHookers->HOOK_ExecCall = MyExecCall;
      pEo->pHookers->HOOK_ExecReturn = MyExecReturn;
    }
    return 0;
  case PreprocessorExeFinish:
    break;
  case PreprocessorExeNoRun:
    break;

    default: /* in any cases that are not handled by the preprocessor just go on */
      printf("Preprocessor invoked, code=%d\n",*pCmd);
      *pCmd = PreprocessorContinue;
      return 0;
    }

  }
