; scriba.conf
; ScriptBasic sample configuration file
;
; Note that this configuration file format is from v1.0b19 or later and has to be compiled
; to internal binary format before starting ScriptBasic

; this is the extension of the dynamic load libraries on this system
dll ".so"

; where the modules are to be loaded from
module "/usr/local/lib/scriba/"

; where to search system and module include files
; trailing / or / is needed
include "/usr/share/scriba/include/"

;
; define external preprocessors
;
preproc (
  internal (
    dbg "/usr/local/lib/scriba/dbg.so"
    )

; extensions that preprocessors are to be applied on
  extensions (
; here the key is the extension and the value is the symbolic name of the external preprocessor
     heb "heb"
     )
; the external preprocessors
  external (
    heb (
      executable "/usr/bin/scriba /usr/share/scriba/source/heber.bas"
      directory "/var/cache/scriba/hebtemp/"
      )
    )
  )

;
; LIMIT VALUES TO STOP INIFINITE LOOP
;

; the maximal number of steps allowed for a program to run
; comment it out or set to zero to have no limit
maxstep 0

; the maximal number of steps allowed for a program to run
; inside a function.
; comment it out or set to zero to have no limit
maxlocalstep 0

; the maximal number of recursive function call deepness
; essentially this is the "stack" size
maxlevel 0


; the maximal memory in bytes that a basic program is allowed to use
; for its variables
maxmem 1000000

;
; ScriptBasic loads the modules before starting the code
; in the order they are specified here
;
;preload "ext_trial"

;
; This is the directory where we store the compiled code
; to automatically avoid recompilation
;
cache "/var/cache/scriba/cache/"

cgi (
;
; These are the keys used by the CGI module
;
  debugfile "/etc/scriba/cgidebug.txt"
  )

;
; berkeley db config
;
bdb (

 ; directories where to store the 
 dir (
   home "/var/cache/scriba/berkeleydb/" ; the home directory of operation of the Berkerley DB
   data "db"  ; database files
   log  "log" ; log files
   temp "tmp" ; temporary files
   )

; ScriptBasic extension modules can access only string configuration values
; therefore you have to specify these numbers within quotes (later versions will change it)
;  limits (
;    lg_max "1024000"
;    mp_mmapsize "0000"
;    mp_size "000"
;    tx_max "000"
;    lk_max "000"
;    )

  ; what lock strategy to use
  ; it can be any of the followings
  ; default oldest random youngest
  lockstrategy default

  flags (
    ; set the value to yes or no
    lock_default no
    init_lock yes
    init_log yes
    init_mpool yes
    init_txn yes
    create yes
    )
  )
;
; MySQL configuration
;
mysql (
  connections (
    test (        ; the name of the connection
	  host "127.0.0.1" ; the host for the connection
	  db "test"   ; database for the connection
	  user "root" ; user for the connection
	  password "" ; password for the connection
	  port 0      ; the port to use
	  socket ""  ; the name of the socket or ""
	  flag 0      ; the client flag
	  clients 10  ; how many clients to serve before really closing the connections
	  )
    auth  (
	  host "127.0.0.1" ; the host for the connection
	  db "auth"   ; database for the connection
	  user "root" ; user for the connection
	  password "" ; password for the connection
	  port 0      ; the port to use
	  socket ""   ; the name of the socket or ""
	  flag 0      ; the client flag
	  clients 10  ; how many clients to serve before really closing the connections
          )
    )
  )
;
; Configure the simple ScriptBasic httpd daemon
;
httpd (
  threads 20
  listenbacklog 30
  port 8080
  home "/home/httpd/cgi-bin/"
  proxyip 0 ; set it true if you use Eszter engine behind Apache rewrite and proxy modules

  pid (
    file "/var/log/scriba/pid.txt"
    delay 10 ; number of seconds to sleep between the examination of the pid file if that still exists
    wait ( 
      period 10 ; number of times to wait for the threads to stop before forcefully exiting
      length 1  ; number of seconds to wait for the threads to stop (total wait= period*length)
      )
    )

  vdirs (
    dir "/cgi-bin/:/home/verhas/scribas/"
    dir "/sibawa/:/home/verhas/sb/ScriptBasic/source/examples/sibawa/cgi-bin/"
    )
  client (
;    allowed "127.0.0.1/255.255.255.255"
    allowed "0.0.0.0/0.0.0.0"

;    denied "127.0.0.1/0.0.0.0"
;    denied "16.192.0.0/255.255.0.0"
    )
  errmsgdest 3
  nolog 0 ; set this true not to use logs or ignore erroneouslog configuration
  run (
    start "/home/verhas/sb/ScriptBasic/source/examples/sibawa/sibawastart.bas"
      )
  log (
    panic ( file "/var/log/scriba/panic.log" )
    app   ( file "/var/log/scriba/app.log" )
    err   ( file "/var/log/scriba/err.log" )
    hit   ( file "/var/log/scriba/hit.log" )
    stat  ( file "/var/log/scriba/stat.log" )
    )
  ; the error page when a page is not found
msg404 """
<HTML>
<HEAD>
<TITLE>Error 404 page not found</TITLE>
</HEAD>
<BODY>
<FONT FACE="Verdana" SIZE="2">
<H1>Page not found</H1>
We regretfully inform you that the page you have requested can not be found on this server.
<p>
In case you are sure that this is a server configuration error, please contact
<FONT SIZE="3"><TT>root@localhost</TT></FONT>
</FONT>
</BODY>
</HTML>
 """
  code404 "200 OK" ; the http error code for page not found. The default is 404
  ; the program to run when a page is not found
  ;run404 "\\MyProjects\\sb\\source\\examples\\run404.bas"
  )
