/*
FILE: httpd.c
HEADER: httpd.h

TO_HEADER:

#ifdef WIN32
typedef u_short port_t;
#else
typedef int port_t;
#define closesocket close
#endif
// AppInit can set the value to different than ...
extern port_t port;
// the default port to listen
#define HTTPD_PORT 9999

// the listen backlog that we pass to the operating system
#define LISTEN_BACKLOG 200
// AppInit can set different listenmax
extern int listenmax;
// AppInit may decide and tell the application under UNIX to do forking
extern int do_fork;

// AppInit can set the value to different than ...
extern int threadmax;
// the maximum number of concurrent hits that we handle
#define CONCURRENT_HITS 100

// Because this http daemon is designed to serve only GET data
// we maximize the size of a http request
#define HIT_MAX_SIZE 32768
// the maximum number of header lines that we can handle
#define MAX_HEADER_LINES 100
// the maximum length of the query string
#define MAX_QUERY_LENGTH 256

struct _fun {
  char *(*pGetServerVariable)();
  int (*pWriteClient)();
  int (*pReadClient)();
  int (*pWriteClientText)();
  int (*pState)();
  int (*pContentType)();
  int (*pHeader)();
  int (*pStartBody)();
  char *(*pGetParam)();
  char *(*pPostParam)();
  char *(*pScriptName)();
  void (*pCloseClient)();
  };
//Defines that help the programmer writing the HttpProc
// make less number of errors
#define GetServerVariable(X) (pT->pFunctions->pGetServerVariable(pT,(X)))
#define WriteClient(X,Y)     (pT->pFunctions->pWriteClient(pT,(X),(Y)))
#define ReadClient(X,Y)      (pT->pFunctions->pReadClient(pT,(X),(Y)))
#define WriteClientText(X)   (pT->pFunctions->pWriteClientText(pT,(X)))
#define State(X)             (pT->pFunctions->pState(pT,(X)))
#define ContentType(X)       (pT->pFunctions->pContentType(pT,(X)))
#define Header(X,Y)          (pT->pFunctions->pHeader(pT,(X),(Y)))
#define StartBody()          (pT->pFunctions->pStartBody(pT))
#define GetParam(X)          (pT->pFunctions->pGetParam(pT,(X)))
#define PostParam(X)         (pT->pFunctions->pPostParam(pT,(X)))
#define ScriptName()         (pT->pFunctions->pScriptName(pT))
#define CloseClient()        (pT->pFunctions->pCloseClient(pT))

typedef struct _ThreadData {
  int ThreadIndex; // the index of this element
  int SocketOpened; // true when the socket is opened
  SOCKET msgsock;// the messaging socket
  int NextFree;  // the next free thread data
  void *pThreadLocalData; // pointer to local data that the thread may use
  void **pApp;   // pointer to the application local data
  THREADHANDLE T;
  struct sockaddr_in addr;

// In these variables we store the header lines so that the application can access it
  int iHeaderLines;            // the number of header lines
  char *HeaderKey[MAX_HEADER_LINES],
       *HeaderString[MAX_HEADER_LINES];

// the string of the method
  char *pszMethod;             // REQUEST_METHOD
// the unaltered query string the way the client has sent to us
  char *pszQueryString;        // QUERY_STRING

  unsigned char ClientIP[4];

  unsigned int  cbAvailable;   // Available number of bytes
  unsigned char *pszData;      // Pointer to cbAvailable bytes
  struct _fun *pFunctions;          // Call Back Function Control Block
  char getparams[MAX_QUERY_LENGTH];
  int getparlen;
  char script[MAX_QUERY_LENGTH];
  int scriptlen;
  char buffer[HIT_MAX_SIZE];
  } ThreadData, *pThreadData;


*/

#include <stdio.h>
#include <string.h>
#ifdef WIN32
#include <winsock2.h>
#include <winbase.h>
#define SLEEPER 1000
#else
#include <sys/time.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>
#define Sleep sleep
#define SLEEPER 1
#endif

#include "thread.h"
#include "filesys.h"
#include "httpd.h"

#define MAXBINDTRIAL 1200 /* so many times do we try the bind*/
#define BINDSLEEP    1   /* so many seconds we sleep after an unsuccessful bind */

/* this function is to be implemented in a separate file and to be linked
   with this framework */
void HttpProc(pThreadData);
/* this is called before anything in main. If it returns non-zero then
   main emits an error and stops (On WNT the socket subsytem is initialized
   before calling AppInit) */
int AppInit(int argc, char*argv[], void **pApp);
/* this is called by main after binding on the desired port, and after forking
   to background on UNIX. This function should start all threads, otherwise the
   forking looses all threads except the main thread. The first version of this
   code started the logger threads before the fork and the parent exited with the
   running logger threads whil the child daemon did not run the logger threads. */
int AppStart(void **pApp);

/*

*/
#ifdef WIN32
/* This is a total DUMMY under Windows NT */
void InitSignalHandlers(){}
#else
void SEGV_handle(int i) {
    char fname[2000];
    int pid;
    pid=getpid();
    sprintf(fname,"/home/infrared/adengine/segv/%d",pid);
    fclose(fopen(fname,"w+"));
    sleep(1000000);
}

/*POD
=section InitSignalHandlers
=H Initialize the signal handlers under UNIX

/*FUNCTION*/
void InitSignalHandlers(
  ){
/*noverbatim
CUT*/
  signal(SIGPIPE,SIG_IGN);
  signal(SIGSEGV,SEGV_handle); 
 
  }

#endif

static struct _StateCode {
  int Code;
  char *StateMessage;
  }StateMessages[]={
  200, "OK",
  201, "CREATED",
  202, "Accepted",
  203, "Partial Information",
  204, "No Response",
  301, "Found, but moved",
  302, "Found, but data resides under different URL (add a /)",
  303, "Method",
  304, "Not Modified",
  400, "Bad request",
  401, "Unauthorized",
  402, "PaymentRequired",
  403, "Forbidden",
  404, "Not found",
  500, "Internal Error",
  501, "Not implemented",
  502, "Service temporarily overloaded",
  503, "Gateway timeout ",
  600, "Bad request",
  601, "Not implemented",
  602, "Connection failed",
  603, "Timed out",
  000, NULL
  };

static int unhex(int t){
  if( t >= 'A' && t <= 'F' )return t-'A'+10;
  if( t >= 'a' && t <= 'f' )return t-'a'+10;
  return t-'0';
  }

static int strIcmp(char *a, char *b){
  int ca,cb;

  while( (ca=*a) && (cb=*b) ){           
    if( isupper(ca) )ca = tolower(ca);
    if( isupper(cb) )cb = tolower(cb);
    if( ca != cb )return ca-cb;
    a++,b++;
    }
  return 0;
  }

static char *_GetServerVariable(pThreadData ThisThread,
                            char *VariableName){
  int i;

  for( i = 0 ; i < ThisThread->iHeaderLines ; i++ ){
    if( ! strIcmp(ThisThread->HeaderKey[i],VariableName) )return ThisThread->HeaderString[i];
    }
  return NULL;
  }

static void _CloseClient(pThreadData ThisThread){
  if( ThisThread->SocketOpened ){
    closesocket(ThisThread->msgsock);
    ThisThread->SocketOpened = 0;
    }
  }

static int _WriteClient(pThreadData ThisThread,
                    char *pszBuffer,
                    int cbBuffer){
  int i;
  fd_set writefds;
  struct timeval timeout;

  FD_ZERO(&writefds);
  FD_SET(ThisThread->msgsock,&writefds);
  timeout.tv_sec=60;
  timeout.tv_usec=0;
  i = select(FD_SETSIZE,NULL,&writefds,NULL,&timeout);
  if( i == 0 )return 1;
  return cbBuffer != send(ThisThread->msgsock,pszBuffer,cbBuffer,0);
  }

static int _WriteClientText(pThreadData ThisThread,
                    char *pszBuffer){
  int i;
  fd_set writefds;
  struct timeval timeout;

  FD_ZERO(&writefds);
  FD_SET(ThisThread->msgsock,&writefds);
  timeout.tv_sec=60;
  timeout.tv_usec=0;
  i = select(FD_SETSIZE,NULL,&writefds,NULL,&timeout);
  if( i == 0 )return 1;

  if( pszBuffer == NULL )
    return 4 != send(ThisThread->msgsock,"null",4,0);

  i = strlen(pszBuffer);
  return i != send(ThisThread->msgsock,pszBuffer,i,0);
  }

static int _State(pThreadData ThisThread,
                   int StateCode){
  int i;
  char buffer[80];

  for( i = 0 ; StateMessages[i].StateMessage && 
               StateMessages[i].Code <= StateCode ; i++ ){
    if( StateMessages[i].Code == StateCode ){
      sprintf(buffer,"HTTP/1.0 %d %s\n",StateCode,StateMessages[i].StateMessage);
      return _WriteClientText(ThisThread,buffer);
      }
    }
  sprintf(buffer,"HTTP/1.0 %d\n",StateCode);
  return _WriteClientText(ThisThread,buffer);
  }

static int _ContentType(pThreadData ThisThread,
                        char *pszContentType){
  _WriteClient(ThisThread,"Content-Type: ",14);
  _WriteClientText(ThisThread,pszContentType);
  return _WriteClient(ThisThread,"\n",1);
  }

static int _Header(pThreadData ThisThread,
                   char *pszHeaderKey,
                   char *pszHeaderValue){
  _WriteClientText(ThisThread,pszHeaderKey);
  _WriteClient(ThisThread,": ",2);
  _WriteClientText(ThisThread,pszHeaderValue);
  return _WriteClient(ThisThread,"\n",1);
  }

static int _StartBody(pThreadData ThisThread){
  return _WriteClient(ThisThread,"\n",1);
  }

static int _ReadClient(pThreadData ThisThread,
                   char *pszBuffer,
                   int cbBuffer){
  int i;
  fd_set readfds;
  struct timeval timeout;

  FD_ZERO(&readfds);
  FD_SET(ThisThread->msgsock,&readfds);
  timeout.tv_sec=60;
  timeout.tv_usec=0;
  i = select(FD_SETSIZE,&readfds,NULL,NULL,&timeout);
  if( i == 0 )return 0;

  return recv(ThisThread->msgsock,pszBuffer,cbBuffer,0);
  }

static char *_ScriptName(pThreadData ThisThread){
  int i,j;

  if( ! ThisThread->script[0] ){
    for( i = 0 ; ThisThread->pszQueryString[i] ; i++ )
      if( ThisThread->pszQueryString[i] == '?' )break;
    while( i && ThisThread->pszQueryString[i] != '/' )i--;
    if( ThisThread->pszQueryString[i] == '/' )i++;
    for( j = 0 ; ThisThread->pszQueryString[i] &&
                 ThisThread->pszQueryString[i] != '?' ; i++, j++ ){
      ThisThread->script[j] = ThisThread->pszQueryString[i];
      if( j >= MAX_QUERY_LENGTH ){
        ThisThread->script[0] = (char)0;
        return NULL;
        }
      }
    ThisThread->script[j] = (char)0;
    }
  return ThisThread->script;
  }

static char *_GetParam(pThreadData ThisThread,
                     char *key){
  char *s;
  int i,j;

  if( ! ThisThread->getparams[0] ){
    /* when first called it contains no data */
    s = ThisThread->pszQueryString;
    while( *s && *s != '?' )s++;
    if( ! *s )return NULL;
    s++; /* step over the ? */
    /* check the size before copiing */
    ThisThread->getparlen = strlen(s);
    if( ThisThread->getparlen >= MAX_QUERY_LENGTH )return NULL;
    strcpy(ThisThread->getparams,s);
    s = ThisThread->getparams;
    for( i=j=0 ; ; i++ ){
       s[i] = s[j];
       if( ! s[j] )break;
       if( s[j] == '%' && s[j+1] && s[j+2] ){
         s[i] = unhex(s[j+1])*16+unhex(s[j+2]);
         j += 3;
         }else j++;
       }
    ThisThread->getparlen = i;
    s = ThisThread->getparams;
    while( *s ){
      if( *s == '&' )*s = (char)0;
      s++;
      }
    }
  /* Now we have the get params in ThisThread->getparams */
  s = ThisThread->getparams;
  for( i = 0 ; i < ThisThread->getparlen ; i++ ){
    for( j = 0 ; key[j] && s[i] && s[i] != '=' ; j++, i++ ){
      if( s[i] != key[j] ){
        while( s[i] )i++;
        break;
        }
      }
    if( s[i] )return s+i+1;
    }
  return NULL;
  }

static char *_PostParam(pThreadData ThisThread,
                      char *key){
  return NULL;
  }

/* This is an array of thread data. They are linked together so that we have a list of
   free threads.  */
static pThreadData threads;
/* This variable indexes the first free thread or has the value -1 if there are no free threads */
static int FirstFreeThread;
/* Mutex to lock the variable FirstFreeThread */
static MUTEX mxFirstFreeThread;
/* The port that we are listening on. This is set to default value by main(), but AppInit max change it from config file. */
port_t port;
/* The maximum number of threads that we start. */
int threadmax;
/* The listen backlog passed as parameter to listen. */
int listenmax;
/* Should we do a fork before starting AppStart or not. */
int do_fork;

SOCKET sock; /* This is a global variable so that the httpproc shutdown can close it. */

int main(int argc, char *argv[]){
  int addr_size;
  int i;
  int length;
  struct _fun Functions;
  struct sockaddr_in server;
  void *AppData;
  pThreadData ThisThread;
  int do_safe;
#ifdef WIN32
  WORD wVersionRequested;
  WSADATA wsaData;
  int err;
#else
  int cpid;
#endif
  do_fork = argc > 1 && !strcmp(argv[1],"-start");
  do_safe = argc > 1 && !strcmp(argv[1],"-safe");
  if( do_safe )do_fork = 1;

#ifdef WIN32
  wVersionRequested = MAKEWORD( 2, 2 );
  err = WSAStartup( wVersionRequested, &wsaData );
  if( err != 0 ){
    fprintf(stderr,"Error initializing the Windows Socket subsystem\n");
    exit(1);
    }
#endif
  port=HTTPD_PORT;
  threadmax = CONCURRENT_HITS;
  listenmax = LISTEN_BACKLOG;

  /* init the signal handlerts to SIG_IGN so that a brokern pipe does not kill us */
  InitSignalHandlers();

  /* This is just a pointer that we guarantee not to be altered. */
  AppData = NULL;
  /* AppInit should return zero on success. */
  if( i=AppInit(argc,argv,&AppData) ){
    fprintf(stderr,"AppInit returned %d\n",i);
    exit(i);
    }

  sock = socket(AF_INET, SOCK_STREAM, 0);
  i=1;
  setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *)(&i),
                   sizeof(i));
  if( sock < 0 ){
    fprintf(stderr, "Error at socket");
    exit(1);
    }
  server.sin_family = AF_INET;   
  server.sin_addr.s_addr = htonl(INADDR_ANY);

  server.sin_port = htons(port);
  for( i=0 ; i<MAXBINDTRIAL ; i++ ){
    if( ! bind(sock, (struct sockaddr *)&server, sizeof(server)) )break;
    if( i == MAXBINDTRIAL-1 ){
      fprintf(stderr, "\nError at bind.");
      exit(1);
      }
    if( i == 0 )fprintf(stderr,"Bind failed, retrying at most %d times\n.",MAXBINDTRIAL);
    else fprintf(stderr,".");
    if( i%40 == 39 )fprintf(stderr,"\n");
    Sleep(BINDSLEEP);
    }
  if( i )fprintf(stderr,"\nBind finally successful after %d trials\n",i);
  
#ifndef WIN32
  setuid(getuid());
  fprintf(stderr,"%d\n",getuid());
/*  fclose(fopen("/etc/eztnemszabad","w+")); */
  
#endif

  length = sizeof(server);
  if( getsockname(sock, (struct sockaddr *)&server, &length) ){ 
    fprintf(stderr, "Error at getsockname.");
    exit(1);
    }
    
  listen(sock, listenmax);

#ifndef WIN32
  /* we have to do the fork here before calling AppInit, because AppInit
     starts threads and fork does NOT duplicate the threads of the parent
     process except the thread calling fork() .*/
  if( do_fork ){
    if( fork() )exit(0);
    }
  if( do_safe ){
    while( cpid = fork() ){
      waitpid( cpid , NULL , WUNTRACED );
      }
    }
#endif

    fprintf(stderr, "calling appstart"); fflush(stderr);
  if( i=AppStart(&AppData) ){
    fprintf(stderr,"Appstart returned %d\n",i);
    exit(i);
    }
fprintf(stderr, " appstart returned"); fflush(stderr);

  threads = (pThreadData)malloc(threadmax*sizeof(ThreadData));
  if( threads == NULL ){
    fprintf(stderr,"Not enough memory\n");
    exit(1);
    }

  /* Initialize the list of thread data
  */
  for( i=0 ; i < threadmax ; i++ ){
    threads[i].ThreadIndex = i;
    threads[i].pFunctions = &Functions;
    threads[i].pApp       = &AppData;
    threads[i].NextFree   = i+1;
    }
  /* make it dead end */
  threads[threadmax-1].NextFree = -1;
  FirstFreeThread = 0;
  thread_InitMutex(&mxFirstFreeThread);
  Functions.pGetServerVariable = _GetServerVariable;
  Functions.pWriteClient       = _WriteClient;
  Functions.pWriteClientText   = _WriteClientText;
  Functions.pReadClient        = _ReadClient;
  Functions.pState             = _State;
  Functions.pContentType       = _ContentType;
  Functions.pHeader            = _Header;
  Functions.pStartBody         = _StartBody;
  Functions.pGetParam          = _GetParam;
  Functions.pPostParam         = _PostParam;
  Functions.pCloseClient       = _CloseClient;
  Functions.pScriptName        = _ScriptName;

  while(1){
    ThisThread = GetFreeThread();
    do{
      addr_size=sizeof(struct sockaddr);
      ThisThread->msgsock = accept(sock,
                                   (struct sockaddr *)&(ThisThread->addr),
                                   &addr_size);
       }while(
#ifdef WIN32
               ThisThread->msgsock == INVALID_SOCKET
#else
               ThisThread->msgsock <= 0
#endif
               );
    ThisThread->SocketOpened = 1;
    thread_CreateThread(&(ThisThread->T),HitHandler,ThisThread);
    }
  /* we never get here */
  return 0;
  }

/*FUNCTION*/
pThreadData GetFreeThread(void
  ){
  pThreadData t;
  /* get exclusive access to the first free thread index */
  thread_LockMutex(&mxFirstFreeThread);
  /* while there is no free thread wait a second and try again */
  while( FirstFreeThread == -1 ){
    thread_UnlockMutex(&mxFirstFreeThread);
    Sleep(1*SLEEPER);
    thread_LockMutex(&mxFirstFreeThread);
    }
  t = threads+FirstFreeThread;
  FirstFreeThread = threads[FirstFreeThread].NextFree;
  thread_UnlockMutex(&mxFirstFreeThread);

  return t;
  }

/*FUNCTION*/
void ReleaseThreadData(int index
  ){
  thread_LockMutex(&mxFirstFreeThread);
  threads[index].NextFree = FirstFreeThread;
  FirstFreeThread = index;
  thread_UnlockMutex(&mxFirstFreeThread);
  }

#define MSPACE(x) while( *x && isspace(*x)) x++
#define SSPACE(x) if( *x && isspace(*x)) x++


/*FUNCTION*/
void HitHandler(void *t
  ){
  pThreadData ThisThread;
  int j,i,cbBuffer,cbCharsRead,cbCharsReadTotal;
  char *Buffer;
  char *s;
  void MyHttpExtensionProc(pThreadData);
  fd_set readfds;
  struct timeval timeout;

#define ABORT_CONNECTION goto FINISH_CONNECTION;

  ThisThread = (pThreadData)t;

  Buffer = ThisThread->buffer;
  cbBuffer = HIT_MAX_SIZE;

#ifdef WIN32
  ThisThread->ClientIP[0] = ThisThread->addr.sin_addr.S_un.S_un_b.s_b1;
  ThisThread->ClientIP[1] = ThisThread->addr.sin_addr.S_un.S_un_b.s_b2;
  ThisThread->ClientIP[2] = ThisThread->addr.sin_addr.S_un.S_un_b.s_b3;
  ThisThread->ClientIP[3] = ThisThread->addr.sin_addr.S_un.S_un_b.s_b4;
#else
  memcpy(ThisThread->ClientIP,&(ThisThread->addr.sin_addr.s_addr),4);
#endif

  cbCharsReadTotal = 0;
  while(1){
    FD_ZERO(&readfds);
    FD_SET(ThisThread->msgsock,&readfds);
    timeout.tv_sec=60;
    timeout.tv_usec=0;
    i = select(FD_SETSIZE,&readfds,NULL,NULL,&timeout);
    if( i == 0 )ABORT_CONNECTION;
    cbCharsRead = recv(ThisThread->msgsock,Buffer,cbBuffer,0);
    /* some clients send such packets, but I could not figure out how */
    if( cbCharsRead == 0 )ABORT_CONNECTION
    if( cbCharsRead < 0 )ABORT_CONNECTION

    /* If this is the first chunk read then we are searching for the header terminating
       empty line from the first character. If this is not the first chunk then we start the
       search from the end of the previous chunk. */
    if( cbCharsReadTotal < 3 )j = 4 - cbCharsReadTotal; else j = 1;

    cbCharsReadTotal += cbCharsRead;
    /* if we have found the empty line following the header */
    for( ; j <= cbCharsRead ; j++ ){
      if( Buffer[j-1] == '\n' && Buffer[j-2] == '\r' &&
          Buffer[j-3] == '\n' && Buffer[j-4] == '\r' ){
        Buffer[j-4] = (char)0;
        ThisThread->pszData = Buffer+j;
        ThisThread->cbAvailable = cbCharsRead - j;
        goto HEADER_IS_READ;
        }
      }
    cbBuffer -= i;
    Buffer += i;
    if( cbBuffer <= 0 )ABORT_CONNECTION
    }
HEADER_IS_READ:;

  /* Here we have the total HTTP GET header. */
  s = ThisThread->pszMethod = ThisThread->buffer;
  while( *s && ! isspace(*s) )s++;
  if( *s )
    *s++ = (char)0;
  else ABORT_CONNECTION

  ThisThread->pszQueryString = s;
  while( *s && ! isspace(*s) )s++;
  if( *s )
    *s++ = (char)0;
  else ABORT_CONNECTION

  /* if there is anything after the URL until the new line then skip it */
  while( *s && *s != '\n' )s++;
  /* skip the CR/LF */
  while( *s == '\n' || *s == '\r' )s++;

  ThisThread->iHeaderLines = 0;
  /* now s points to the start of the first header line */
  while( *s ){

    ThisThread->HeaderKey[ThisThread->iHeaderLines] = s;
    while( *s && ! isspace(*s) && *s != ':' )s++;
    if( *s )*s++ = (char)0;
    while( isspace(*s) )s++;
    if( *s ){
      ThisThread->HeaderString[ThisThread->iHeaderLines] = s;
      while( *s && *s != '\n' && *s != '\r' )s++;
      if( *s )*s++ = (char)0;
      while( *s == '\n' || *s == '\r' )s++;
      ThisThread->iHeaderLines++;
      }
    }

  ThisThread->getparams[0] = (char)0;
  ThisThread->script[0] = (char)0;
  ThisThread->pThreadLocalData = NULL;

  HttpProc(ThisThread);

FINISH_CONNECTION:
  if( ThisThread->SocketOpened ){
    closesocket(ThisThread->msgsock);
    ThisThread->SocketOpened = 0;
    }
  ReleaseThreadData(ThisThread->ThreadIndex);
  thread_ExitThread();
  }
